/*
 * 18Fxxx BOOT LOADER Copyright (C) 2002 HI-TECH Software
 *	This software is freely distributable and may be used
 *	for any purpose. No warranty of any kind is provided
 *	and all use is entirely at your own risk.

 *	Test with Hi_tec corrected code 4.11.2002 / JNi
 *	Code modified for DIS prototype board 18.12.2002	
 */


#include <pic18.h>
#include "disboot.h"

	__CONFIG(1, HS);		// config bits for 18F452 / v8.20
	__CONFIG(2, PWRTEN & WDTDIS);
	__CONFIG(4, LVPDIS);
//	__CONFIG(5,CPB);

near unsigned char buff[8];	/* 8 byte buffer used for writing to program memory */
near unsigned char DO_NOT_INCREMENT;
near unsigned char cksum,rectype,bcount;
near unsigned char delay_time;
near unsigned char index;
near unsigned short erase;


#ifdef VERBOSE
unsigned char START_MSG[]="JNi Loader 2002\n";
unsigned char DOWNLOAD_MSG[]="\rDownload-";
#endif


#asm
psect intcode	// interrupts jump to new routine addresses
	goto	PROG_START+8

psect intcodelo
	goto	PROG_START+0x18

global clear_common	// do not want the following startup routines
clear_common:		// defines these as blank
global clear_ram
clear_ram:
#endasm


/* PUTCH() - outputs 1 byte to the serial port */

#ifdef VERBOSE
void putch(unsigned char byte) 

{
	while(!TXIF);	/* set when register is empty */
	TXREG = byte;	/* output one byte */
}

/* Output a string via the serial port */

void puts(const char *s)

{				
	while(s && *s)		
		putch(*s++);	
}
#endif

/* Get a 4 bit HEX number from the serial port */

unsigned char gx(void)
{
	while(!RCIF);
	EEDATA=RCREG;
#ifdef VERBOSE
	putch(EEDATA);	// echo input nibbles to output
#endif
	if(EEDATA>='A')
		return ((EEDATA - (unsigned char)'A')+10);
	return (EEDATA - '0');
}


/* Get an 8 bit HEX number from the serial port */

unsigned char g2x(void)
{
	return((gx()<<4)+gx());
}


/* check the checksum */

void checksum(void)
{
	if(cksum += g2x())	// if checksum does not add to zero, bad check, reset
		RESET();
	#ifdef VERBOSE
	putch('\r');	// echo each hex record on a new line
	#endif
}
	

/* Initiate a write to memory */

void zap(void)
{
	WREN=1;
	EECON2=0x55;
	EECON2=0xAA;
	WR=1;
	NOP();
	while(WR);
	WREN=0;
}


/* transfer a byte from RAM to internal flash register(s) */

void table_write(void)

{
	if(DO_NOT_INCREMENT)	// address is already loaded in TBLPTR, no pre-increment needed
		asm("tblwt*");
	else
		asm("tblwt+*");	// otherwise TBLPTR must be pre-incremented
	DO_NOT_INCREMENT=0;
}



/* write the 8 byte buffer to flash */

void flash8()
{
	RTS = 1;				// stop PC transmitter
	LEDB = !LEDB;

	if(DO_NOT_INCREMENT)
		TBLPTRL&=0xF8;	// point to start of 8 byte panel
	for(index=0;index<8;)
	{	
		TABLAT = buff[index++];
		table_write();
	}
	zap();

	RTS = 0;				// enable PC transmitter
}


void clear_buffer(void)
{
	buff[0] = buff[1] = buff[2] = FILL_BYTE;	// 8 byte buffer initialised with	
	buff[3] = buff[4] = buff[5] = FILL_BYTE;	// known data
	buff[6] = buff[7] = FILL_BYTE;
}

main()
{
#if defined(_18F242) || defined(_18F252) || defined(_18F442) || defined(_18F452)
	PIE1=0;
#endif
	TRISC = TRISC_INIT;
	TRISE = TRISE_INIT;	// set up led & control port
	LEDB = 1;		// boot mpde indigator
	init_comms();		// sets up the serial port for communication

/* Bootloader waits for a specified time. If the serial port gets no response in */
/* this time, execution of existing program begins. */
#ifdef VERBOSE
	puts(START_MSG);
#endif

// Jumper J26 selected download

	if (J26)					// J28 open
		(*((void(*)(void))PROG_START))(); 	// start user code

	LEDA = 1;		// Red light on. Wait for J26 opening.
	while(!J26)
		;		// wait for jumper release and start loader


//	for(delay_time=BOOT_TIMEOUT; delay_time; --delay_time)
//	{
//		if (RCIF)
//			break;
//#ifdef VERBOSE	/* display a countdown for user response */
//		puts(DOWNLOAD_MSG);
//		putch('0'+delay_time);
//#else
//		TXREG=('0'+delay_time);
//		while(!TXIF);
//		TXREG='\r';
//#endif
//		for(erase=0xFFFF;erase;erase--)	// adjust value for precise timing
//		{
//			INTCON3=0;	// the instructions here are to add delay time
//			PIE2=0;		// to the for loop. They also serve the double
//			INTCON=0;	// purpose of disabling the interrupts.
//		}
//	}
//	if (!RCIF)	/* if no hex file to download, resume normal program */
//	{
//		(*((void(*)(void))PROG_START))();
//	}


	TXREG=':';

	TBLPTRU=0;	/* Clear program memory prior to program download */
	erase=PROG_START;

	while(1)
	{
		TBLPTR=(far unsigned char *)erase;
		EECON1=0x90;	// set up for flash erasure
		zap();
		erase+=64;
#ifdef UPPER_ADDRESS_BYTE
		if(erase==0)
			TBLPTRU++;
		if(TBLPTRU==UPPER_ADDRESS_BYTE)
#else
		if(erase==MEM_TOP)
#endif
			break;
	}

	LEDA = 0;		// ready to start

#ifdef UPPER_ADDRESS_BYTE
	TBLPTRU=0;
#endif

/* receive a hex file via the serial port and write it to program memory */

	for(;;)		// loop until end of file
	{
		while (RCREG!=':');		// wait for start of hex file line 
#ifdef VERBOSE
		putch(':');
#endif		
		cksum = bcount = g2x();		// get the byte count
 
#if defined(EEADRH)
		EEADRH = g2x();
		cksum += TBLPTRH = EEADRH;		// get the address
#else
		cksum += TBLPTRH = g2x();
#endif

		cksum += TBLPTRL = EEADR = g2x();
		DO_NOT_INCREMENT = 1;
		cksum += rectype = g2x();	// get the record type 

		switch(rectype)
		{
			case 0x00:	// data record 
#ifdef VERBOSE

				if((FLASH)&&(TBLPTRU==0)&&(TBLPTRH<3))
					break;
#endif
			clear_buffer();
				while(bcount--)
				{	
					TABLAT = EEDATA = buff[(EEADR&7)] = g2x();	// get the data
					cksum += TABLAT;

					if((CONFIG)||(EEPROM))
					{
						if(CONFIG)	// EEPROM/config. bytes are written one byte at a time
							table_write();

						RTS = 1;	// stop PC transmitter
						zap();
						RTS = 0;
					}
					else
						if((EEADR&7)==7)	// program/IDLOCs are flashed 8 bytes at a time
						{
							flash8();
							clear_buffer();
						}
					EEADR++;
				}	
				if(((EEADR&7)!=0)&&(FLASH))
                                        flash8();

				checksum();
				break;

			case 0x01:	// end of hex file

				checksum();
				TXREG=')';
				
				LEDC = 1;			// Green light on
				for(;;)				// and wait here for Reset
					;

//				(*((void(*)(void))PROG_START))(); // jump to new program 
//				break;


			case 0x04:	// extended address record
				while(bcount--)
				{
					EEADR=g2x();	// this byte determines whether EE, Config or ID data
					cksum += EEADR;
				}

				EEPGD=1;
				if(EEADR==0xF0)
					EEPGD=0;	// select for EEPROM

				CFGS=0;
				if((EEADR&0xF0)==0x30)
					CFGS=1;		// select for config write

				TBLPTRU=EEADR;
				checksum();
				break;

		}
	}
}



